/* This file is part of the ili2fme project.
 * For more information, please see <http://www.interlis.ch/>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package ch.interlis.ili2fme;

import java.util.ArrayList;
import java.util.List;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;

import COM.safe.fme.pluginbuilder.IFMEReader;
import COM.safe.fmeobjects.IFMEFeature;
import COM.safe.fmeobjects.IFMEFeatureVector;
import COM.safe.fme.pluginbuilder.IFMEMappingFile;
import COM.safe.fmeobjects.IFMESession;
import COM.safe.fmeobjects.FMEException;
import COM.safe.fmeobjects.IFMELogFile;

import ch.ehi.basics.logging.EhiLogger;

import ch.interlis.ili2c.metamodel.*;

//import ch.interlis.iom.swig.iom_javaConstants;
import ch.interlis.iom.IomConstants;
import ch.interlis.iom.*;
import ch.interlis.iom_c.Iom;
import ch.interlis.iox.*;
import ch.interlis.iom_j.itf.ItfReader;

/** INTERLIS implementation of an FME-Reader.
 * @author ce
 */
public class Ili2Reader implements IFMEReader {
	private IFMEMappingFile mappingFile=null;
	private IFMESession session=null;
	private String readerKeyword=null;
	private ch.interlis.ili2c.metamodel.TransferDescription iliTd=null;
	private ArrayList iliModelv=null;
	private HashMap tag2class=null; // map<String iliQName,Viewable|AttributeDef modelele>
	private Iterator fmeFeatureTypei=null;
	private HashMap fmeFeaturTypev=null; // map<String fmeFeatureTypeName, ViewableWrapper>
	private String xtfFile=null; // null if ili-file given
	private IoxReader ioxReader=null;
	private java.io.FileInputStream inputFile=null; // input stream to ioxReader
	private IomFile iomFile=null;
	private String currentBid=null;
	private boolean surfaceAsPolyline=false;
	private int iliQNameSize=0;
	private int formatMode=0;
	private static final int MODE_XTF=1;
	private static final int MODE_ITF=2;
	public Ili2Reader(IFMESession session1,IFMEMappingFile mappingFile1,String keyword,IFMELogFile log){
		mappingFile=mappingFile1;
		readerKeyword=keyword;
		session=session1;
		listener=Ili2fme.setupLogging(log);
	}
	public void open(ArrayList args) throws Exception {
		if(args.size()==0){
			throw new IllegalArgumentException("args.size()==0");
		}
		String models=null;
		String modeldir=null;
		for(int i=0;i<args.size();i++){
			String arg=(String)args.get(i);
			EhiLogger.traceState("arg["+Integer.toString(i)+"] "+arg);
			if(arg.equals(Ili2fme.MODELS)){
				i++;
				models=(String)args.get(i);
			}else if(arg.equals(Ili2fme.MODEL_DIR)){
				i++;
				modeldir=(String)args.get(i);
			}else if(arg.startsWith("+")){
				i++; // skip this and next(!) argument
			}else{
				// skip this argument
			}
		}
		Iterator elei=mappingFile.elements();
		while(elei.hasNext()){
			Object eleo=elei.next();
			EhiLogger.traceState("element "+eleo.getClass()+","+eleo);
			if(eleo instanceof ArrayList){
				ArrayList ele=(ArrayList)eleo;
				String val=(String)ele.get(0);
				if(val.equals(readerKeyword+"_"+Ili2fme.MODELS)){
					models=(String)ele.get(1);	
				}else if(val.equals(readerKeyword+"_"+Ili2fme.MODEL_DIR)){
					modeldir=(String)ele.get(1);	
				}
			}
		}
		
		if(models==null){
			//models=mappingFile.fetchString(readerKeyword+"_"+Ili2fme.MODELS);
			models="XTF";
		}
		EhiLogger.traceState("models <"+models+">");
		
		xtfFile=(String)args.get(0);
		if(xtfFile.length()>=2 && xtfFile.charAt(0)=='/' && xtfFile.charAt(1)=='/'){
			StringBuffer x=new StringBuffer(xtfFile);
			x.setCharAt(0,'\\');
			x.setCharAt(1,'\\');
			xtfFile=x.toString();
		}
		EhiLogger.logState("xtfFile <"+xtfFile+">");

		if(modeldir==null){
			//modeldir=mappingFile.fetchString(readerKeyword+"_"+Ili2fme.MODEL_DIR);
			modeldir=new java.io.File(session.fmeHome(),"plugins/interlis2/ilimodels").getAbsolutePath();
			modeldir=new java.io.File(xtfFile).getParentFile().getAbsolutePath()+";"+modeldir;
		}else{
			int startPos=modeldir.indexOf(Ili2fme.XTFDIR_PLACEHOLDER);
			if(startPos>-1){
				StringBuffer buf=new StringBuffer(modeldir);
				buf.replace(startPos,startPos+Ili2fme.XTFDIR_PLACEHOLDER.length(),new java.io.File(xtfFile).getParentFile().getAbsolutePath());
				modeldir=buf.toString();
			}
		}
		EhiLogger.traceState("modeldir <"+modeldir+">");
		
		String xtfExt=ch.ehi.basics.view.GenericFileFilter.getFileExtension(new java.io.File(xtfFile)).toLowerCase();
		if(xtfExt.equals("itf")){
			formatMode=MODE_ITF;
		}else{
			formatMode=MODE_XTF;
		}
		if(xtfExt.equals("ili")){
			ArrayList modeldirv=new ArrayList();
			modeldirv=new ArrayList(java.util.Arrays.asList(modeldir.split(";")));
			ArrayList iliFilev=new ArrayList();
			iliFilev.add(xtfFile);
			xtfFile=null;
			// compile models
			iliTd=ch.interlis.ili2c.Main.compileIliFiles(iliFilev,modeldirv,null);
			if(iliTd==null){
				// compiler failed
				throw new IllegalArgumentException("INTERLIS compiler failed");
			}
			// INTERLIS 1?
			if(iliTd.getIli1Format()!=null){
				formatMode=MODE_ITF;
			}else{
				formatMode=MODE_XTF;
			}
		}else if(models.equals("XTF")){
			// get model names out of transfer file
			iliModelv=new ArrayList();
			if(formatMode==MODE_XTF){
				setupIom();
				iomFile=Iom.openFile(xtfFile,0,null);
				if(iomFile==null){
					throw new IllegalArgumentException("failed to open input file <"+xtfFile+">");
				}
				ioxReader=new ch.interlis.iom_c.Iom_cIoxReader(iomFile);
			}else if(formatMode==MODE_ITF){
				inputFile=new java.io.FileInputStream(xtfFile);
				ioxReader=new ch.interlis.iom_j.itf.ItfReader(inputFile);
			}else{
				throw new IllegalStateException("unexpected formatMode");
			}

			// get first basket
			IoxEvent event;
			StartBasketEvent be=null;
			do{
				event=ioxReader.read();
				if(event instanceof StartBasketEvent){
					be=(StartBasketEvent)event;
					break;
				}
			}while(!(event instanceof EndTransferEvent));
			ioxReader.close();
			ioxReader=null;
			if(inputFile!=null){
				inputFile.close();
				inputFile=null;
			}
			// no baskets?
			if(be==null){
				// no models!
				iliTd=null;
				throw new IllegalArgumentException("no baskets in xtf-file");
			}
			String model[]=be.getType().split("\\.");
			EhiLogger.logState("model from xtf <"+model[0]+">");
			iliModelv.add(model[0]);
			ArrayList modeldirv=new ArrayList();
			modeldirv=new ArrayList(java.util.Arrays.asList(modeldir.split(";")));
			// compile models
			iliTd=ch.interlis.ili2c.Main.compileIliModels(iliModelv,modeldirv,null);
			if(iliTd==null){
				// compiler failed
				throw new IllegalArgumentException("INTERLIS compiler failed");
			}
		}else{
			ArrayList modeldirv=new ArrayList();
			modeldirv=new ArrayList(java.util.Arrays.asList(modeldir.split(";")));
			// parse string
			iliModelv=new ArrayList(java.util.Arrays.asList(models.split(";")));
			// compile models
			iliTd=ch.interlis.ili2c.Main.compileIliModels(iliModelv,modeldirv,null);
			if(iliTd==null){
				// compiler failed
				return;
			}
		}
		
		if(formatMode==MODE_XTF){
			fmeFeaturTypev=ModelUtility.getXtfTransferViewables(iliTd);
		}else if(formatMode==MODE_ITF){
			fmeFeaturTypev=ModelUtility.getItfTransferViewables(iliTd);
		}else{
			throw new IllegalStateException("unexpected formatMode");
		}
		fmeFeatureTypei=null;
		
		// setup mapping from xml-elementnames to types
		if(formatMode==MODE_XTF){
			tag2class=ch.interlis.ili2c.generator.XSDGenerator.getTagMap(iliTd);
		}else if(formatMode==MODE_ITF){
			tag2class=ch.interlis.iom_j.itf.ModelUtilities.getTagMap(iliTd);
		}else{
			throw new IllegalStateException("unexpected formatMode");
		}
		
		// get longest qualified ili name
		Iterator iliqnamei=tag2class.keySet().iterator();
		while(iliqnamei.hasNext()){
			String iliqname=(String)iliqnamei.next();
			if(iliqname.length()>iliQNameSize){
				iliQNameSize=iliqname.length();
			}
		}
		iliQNameSize+=1;
		
		// ASSERT: ready to scan schema or read data
	}

	// Terminate the reader mid-stream.  Any special actions to shut
	// down a reader not finished reading data should be taken in this
	// method.  For most readers, nothing will be done.
	public void abort() throws Exception {
		cleanup();
	}

	// Close the reader after it has exhausted its data source.
	public void close() throws Exception {
		cleanup();
	}
	
	// Return the unique ID assigned to this reader.
	// All Java reader return 0.
	public int id() {
		return 0;
	}

	public IFMEFeature read(IFMEFeature ret) 
		throws Exception 
	{
		// ili-file mode?
		if(xtfFile==null){
			// no features, just schema features!
			return null;
		}
		try{
			// if first call?
			if(ioxReader==null){
				if(formatMode==MODE_XTF){
					if(iomFile==null){
						setupIom();
						iomFile=Iom.openFile(xtfFile,0,null);
						if(iomFile==null){
							throw new IllegalArgumentException("failed to open input file <"+xtfFile+">");
						}
					}
					ioxReader=new ch.interlis.iom_c.Iom_cIoxReader(iomFile);
				}else if(formatMode==MODE_ITF){
					inputFile=new java.io.FileInputStream(xtfFile);
					ioxReader=new ch.interlis.iom_j.itf.ItfReader(inputFile);
					((ItfReader)ioxReader).setModel(iliTd);					
				}else{
					throw new IllegalStateException("unexpected formatMode");
				}
			}
			IoxEvent event;
			while(true){
				event=ioxReader.read();
				//EhiLogger.debug("event "+event.getClass().getName());
				if(event instanceof ObjectEvent){
					ObjectEvent oe=(ObjectEvent)event;
					IomObject iomObj=oe.getIomObject();
					// translate object
					try{
						//EhiLogger.debug("object "+iomObj.getobjecttag()+" "+iomObj.getobjectoid());
						mapFeature(ret,iomObj,null);
					}catch(InputDataException ex){
						EhiLogger.logError("object "+iomObj.getobjecttag()+" "+iomObj.getobjectoid()+" skipped; "+ch.interlis.iom_j.Iom_jObject.dumpObject(iomObj),ex);
						ret.dispose();
						ret=session.createFeature();
						continue;
						//return null;
					}catch(java.lang.Exception ex){
						EhiLogger.logError("object "+iomObj.getobjecttag()+" "+iomObj.getobjectoid()+" "+ch.interlis.iom_j.Iom_jObject.dumpObject(iomObj),ex);
						return null;
					}
					return ret;
				}else if(event instanceof StartBasketEvent){
					StartBasketEvent be=(StartBasketEvent)event;
					currentBid=be.getBid();
					// map basket (to a feature! to get metainfo about basket)
					EhiLogger.logState(be.getType()+" "+currentBid+"...");
					mapBasket(ret,be);
					return ret;
				}else if(event instanceof EndBasketEvent){
					currentBid=null;
				}else if(event instanceof EndTransferEvent){
					// end of file reached
					ioxReader.close();
					ioxReader=null;
					if(inputFile!=null){
						inputFile.close();
						inputFile=null;
					}
					if(listener.hasSeenErrors()){
						throw new Exception("INTERLIS 2 reader failed");
					}
					return null;
				}else{
					// ignore other events
				}
			}
		}catch(Exception ex){
			EhiLogger.logError(ex);
			throw ex;
		}
	}
	private void mapBasket(IFMEFeature ret,StartBasketEvent be)
	{
		ret.setFeatureType(Ili2fme.XTF_BASKETS);
		ret.setStringAttribute(Ili2fme.XTF_TOPIC,be.getType());
		ret.setStringAttribute(Ili2fme.XTF_ID,be.getBid());
	}
	private void mapFeature(IFMEFeature ret,IomObject iomObj,String prefix)
	throws InputDataException
	{
		boolean isStruct=prefix!=null;
		if(!isStruct)prefix="";
		String tag=iomObj.getobjecttag();
		ViewableWrapper wrapper=null;
		if(formatMode==MODE_XTF){
			Viewable aclass=(Viewable)tag2class.get(tag);
			if(aclass==null){
				EhiLogger.logError("line "+iomObj.getobjectline()+": unknonw class <"+tag+">; ignored");
				return;
			}
			Viewable rootClass=ModelUtility.getRoot(aclass);
			if(rootClass==null){
			 rootClass=aclass;
			}
			wrapper=((ViewableWrapper)fmeFeaturTypev.get(rootClass.getScopedName(null)));
		}else if(formatMode==MODE_ITF){
			wrapper=((ViewableWrapper)fmeFeaturTypev.get(tag));
			if(wrapper==null){
				EhiLogger.logError("line "+iomObj.getobjectline()+": unknonw class <"+tag+">; ignored");
				return;
			}
		}
		//EhiLogger.debug("aclass <"+aclass+">, root <"+rootClass+">");
		if(!isStruct){
			ret.setFeatureType(wrapper.getFmeFeatureType());
			ret.setStringAttribute(Ili2fme.XTF_ID,iomObj.getobjectoid());
			ret.setStringAttribute(Ili2fme.XTF_BASKET,currentBid);
		}
		if(formatMode==MODE_ITF){
			// if SURFACE helper table
			if(wrapper.getViewable()==null && wrapper.getSurfaceOrAreaAttr().getDomainResolvingAliases() instanceof SurfaceType){
				//add ref to main table
				String fkName=wrapper.getSurfaceOrAreaAttr().getContainer().getName();
				IomObject structvalue=iomObj.getattrobj(fkName,0);
				String refoid=structvalue.getobjectrefoid();
				ret.setStringAttribute(fkName,refoid);
			}
		}
		AttributeDef geomattr=isStruct ? null : wrapper.getGeomAttr();
		if(geomattr!=null){
			ret.setStringAttribute(Ili2fme.XTF_GEOMATTR,geomattr.getName());
		}
		 ret.setStringAttribute(prefix+Ili2fme.XTF_CLASS,tag);
		Iterator iter;
		if(formatMode==MODE_ITF){
			iter = wrapper.getAttrIterator();
		}else{
			Viewable aclass=(Viewable)tag2class.get(tag);
			iter = aclass.getAttributesAndRoles2();
		}

		boolean geometryDone=false;
		while (iter.hasNext()) {
			ViewableTransferElement prop = (ViewableTransferElement)iter.next();
			if (prop.obj instanceof AttributeDef) {
				AttributeDef attr = (AttributeDef) prop.obj;
				Type type = attr.getDomainResolvingAliases();
				 String attrName=attr.getName();
				if (type instanceof CompositionType){
				 int valuec=iomObj.getattrvaluecount(attrName);
				 for(int valuei=0;valuei<valuec;valuei++){
					IomObject value=iomObj.getattrobj(attrName,valuei);
					if(value!=null){
						mapFeature(ret,value,attr.getName()+"{"+Integer.toString(valuei)+"}.");
					}
				 }
				 // enqueue struct values
				 //int structc=iomObj.getattrvaluecount(attrName);
				 //for(int structi=0;structi<structc;structi++){
				 //	IomObject struct=iomObj.getattrobj(attrName,structi);
			 	 //	enqueStructValue(sqlId,sqlType,attrName,struct,structi);
				 //}
				}else if (type instanceof PolylineType){
					 if(attr!=geomattr){
						 continue;
					 }
					 geometryDone=true;
					 //ret.setStringAttribute("fme_geometry{0}", "xtf_polyline");
					 IomObject value=iomObj.getattrobj(attrName,0);
					 if(value!=null){
						 ret.setStringAttribute(Ili2fme.XTF_GEOMTYPE,"xtf_polyline");
						 setPolyline(ret,value,false,getP((PolylineType)type));
					 }					
				 }else if(type instanceof SurfaceType){
					if(attr!=geomattr){
						 continue;
					 }
					 // is itf and surface attr in main table?
					 if(formatMode==MODE_ITF && wrapper.getViewable()!=null){
					 	// no geometry
					 	continue;
					 }
					 geometryDone=true;
					 //ret.setStringAttribute("fme_geometry{0}", "xtf_surface");
					 IomObject value=iomObj.getattrobj(attrName,0);
					 if(value!=null){
						if(formatMode==MODE_ITF){
							ret.setStringAttribute(Ili2fme.XTF_GEOMTYPE,"xtf_polyline");
							setPolyline(ret,value,false,getP((SurfaceType)type));
						}else{
							if(surfaceAsPolyline){
								ret.setStringAttribute(Ili2fme.XTF_GEOMTYPE,"xtf_polyline");
							}else{
								ret.setStringAttribute(Ili2fme.XTF_GEOMTYPE,"xtf_surface");
							}
							setSurface(ret,value,(SurfaceType)type);
						}
					 }
				 }else if(type instanceof AreaType){
					if(attr!=geomattr){
						 continue;
					 }
					 geometryDone=true;
					 //ret.setStringAttribute("fme_geometry{0}", "xtf_area");
					 IomObject value=iomObj.getattrobj(attrName,0);
					 if(value!=null){
						if(formatMode==MODE_ITF){
							// is helper table?
							if(wrapper.getViewable()==null){
								ret.setStringAttribute(Ili2fme.XTF_GEOMTYPE,"xtf_polyline");
								setPolyline(ret,value,false,getP((AreaType)type));
							}else{
								// main table
								ret.setStringAttribute(Ili2fme.XTF_GEOMTYPE,"xtf_coord");
								ret.setGeometryType(IFMEFeature.FME_GEOM_POINT);
								addCoord(ret,value);
							}
						}else{
							if(surfaceAsPolyline){
								ret.setStringAttribute(Ili2fme.XTF_GEOMTYPE,"xtf_polyline");
							}else{
								ret.setStringAttribute(Ili2fme.XTF_GEOMTYPE,"xtf_area");
							}
							setSurface(ret,value,(AreaType)type);
						}
					 }
				 }else if(type instanceof CoordType){
					if(attr!=geomattr){
						 continue;
					 }
					 geometryDone=true;
					 //ret.setStringAttribute("fme_geometry{0}", "xtf_coord");
					 IomObject value=iomObj.getattrobj(attrName,0);
					 if(value!=null){
						if(!value.getobjecttag().equals("COORD")){
							throw new InputDataException("COORD expected for attribute "+attrName);
						}
						 ret.setStringAttribute(Ili2fme.XTF_GEOMTYPE,"xtf_coord");
						 ret.setGeometryType(IFMEFeature.FME_GEOM_POINT);
						 addCoord(ret,value);
					 }
				}else{
				 String value=iomObj.getattrvalue(attrName);
				 if(value!=null){
					 ret.setStringAttribute(prefix+attrName, value);
				 }
				}
			}
			if(prop.obj instanceof RoleDef){
				RoleDef role = (RoleDef) prop.obj;
				String roleName=role.getName();
				// a role of an embedded association?
				if(prop.embedded){
					AssociationDef roleOwner = (AssociationDef) role.getContainer();
					if(roleOwner.getDerivedFrom()==null){
						// not just a link?
						 IomObject structvalue=iomObj.getattrobj(roleName,0);
						 if(structvalue!=null){
							if (roleOwner.getAttributes().hasNext()
								|| roleOwner.getLightweightAssociations().iterator().hasNext()) {
								 // TODO add association attributes
							}
							 String refoid=structvalue.getobjectrefoid();
							 long orderPos=structvalue.getobjectreforderpos();
							 if(orderPos!=0){
								// refoid,orderPos
								ret.setStringAttribute(prefix+roleName, refoid);
								ret.setStringAttribute(prefix+roleName+"."+Ili2fme.ORDERPOS, Long.toString(orderPos));
							 }else{
								// refoid
								ret.setStringAttribute(prefix+roleName, refoid);
							 }
						 }
					}
				}else{
				 IomObject structvalue=iomObj.getattrobj(roleName,0);
				 String refoid=structvalue.getobjectrefoid();
				 long orderPos=structvalue.getobjectreforderpos();
				 if(orderPos!=0){
					// refoid,orderPos
					ret.setStringAttribute(prefix+roleName, refoid);
					ret.setStringAttribute(prefix+roleName+"."+Ili2fme.ORDERPOS, Long.toString(orderPos));
				 }else{
					// refoid
					ret.setStringAttribute(prefix+roleName, refoid);
				 }
				}
			}
		}
	}
	private HashMap typeCache=new HashMap();
	private double getP(LineType type)
	{
		if(typeCache.containsKey(type)){
			return ((Double)typeCache.get(type)).doubleValue();
		}
		double p;
		CoordType coordType=(CoordType)type.getControlPointDomain().getType();
		NumericalType dimv[]=coordType.getDimensions();
		int accuracy=((NumericType)dimv[0]).getMaximum().getAccuracy();
		if(accuracy==0){
			p=0.5;
		}else{
			p=Math.pow(10.0,-accuracy);
			//EhiLogger.debug("accuracy "+accuracy+", p "+p);
		}
		typeCache.put(type,new Double(p));
		return p;
	}
	public IFMEFeature readSchema(IFMEFeature ret) throws Exception {
		try{
			// ili2c failed?
			if(iliTd==null){
				return null;
			}
			// first call?
			if(fmeFeatureTypei==null){
				// create XTF_BASKETS class
				ret.setFeatureType(Ili2fme.XTF_BASKETS);
				ret.setStringAttribute("fme_geometry{0}", "xtf_none");
				ret.setStringAttribute(Ili2fme.XTF_TOPIC,getIliQNameType());
				ret.setStringAttribute(Ili2fme.XTF_ID,Ili2fme.ID_TYPE);
				// init class iterator
				fmeFeatureTypei=fmeFeaturTypev.keySet().iterator();
				return ret;
			}
			// more classes?
			while(fmeFeatureTypei.hasNext()){
				String fmeFeatureType=(String)fmeFeatureTypei.next();
				//EhiLogger.debug("fmeFeatureType "+fmeFeatureType);
				ViewableWrapper wrapper=(ViewableWrapper)fmeFeaturTypev.get(fmeFeatureType);
				if(wrapper.getViewable()!=null){
					Viewable v=wrapper.getViewable();
					if((v instanceof Table) && !((Table)v).isIdentifiable()){
						continue;
					}
				}else{
					// if SURFACE helper table
					if(wrapper.getSurfaceOrAreaAttr().getDomainResolvingAliases() instanceof SurfaceType){
						//add FK to main table
						ret.setStringAttribute(wrapper.getSurfaceOrAreaAttr().getContainer().getName(),Ili2fme.ID_TYPE);
					}
					
				}
				ArrayList geomAttrAccu=new ArrayList();
				mapFeatureType(geomAttrAccu,ret,wrapper,null);						
				AttributeDef geomattr=wrapper.getGeomAttr();
				if(geomattr==null){
					ret.setStringAttribute("fme_geometry{0}", "xtf_none");
				}else{
					Type type=geomattr.getDomainResolvingAliases();
					if(formatMode==MODE_XTF){
						if (type instanceof PolylineType){
							ret.setStringAttribute("fme_geometry{0}", "xtf_polyline");
						}else if(type instanceof SurfaceType){
							if(surfaceAsPolyline){
								ret.setStringAttribute("fme_geometry{0}", "xtf_polyline");
							}else{
								ret.setStringAttribute("fme_geometry{0}", "xtf_surface");
							}
						}else if(type instanceof AreaType){
							if(surfaceAsPolyline){
								ret.setStringAttribute("fme_geometry{0}", "xtf_polyline");
							}else{
								ret.setStringAttribute("fme_geometry{0}", "xtf_area");
							}
						}else if(type instanceof CoordType){
							ret.setStringAttribute("fme_geometry{0}", "xtf_coord");
						}else{
							throw new IllegalStateException("!(type instanceof geomType)");
						}
					}else if(formatMode==MODE_ITF){
						if (type instanceof PolylineType){
							ret.setStringAttribute("fme_geometry{0}", "xtf_polyline");
						}else if(type instanceof SurfaceType){
							if(wrapper.getViewable()!=null){
								ret.setStringAttribute("fme_geometry{0}", "xtf_none");
							}else{
								// polyline table
								ret.setStringAttribute("fme_geometry{0}", "xtf_polyline");
							}
						}else if(type instanceof AreaType){
							// if main table?
							if(wrapper.getViewable()!=null){
								ret.setStringAttribute("fme_geometry{0}", "xtf_coord");
							}else{
								// polyline table
								ret.setStringAttribute("fme_geometry{0}", "xtf_polyline");
							}
						}else if(type instanceof CoordType){
							ret.setStringAttribute("fme_geometry{0}", "xtf_coord");
						}else{
							throw new IllegalStateException("!(type instanceof geomType)");
						}
					}
					for(int attri=1;attri<geomAttrAccu.size();attri++){
						geomattr=(AttributeDef)geomAttrAccu.get(attri);
						EhiLogger.logAdaption(geomattr.getScopedName(null)+": Attribute ignored; FME supports only one geometry per feature");							
					}
				}
				return ret;
			}
			return null;
		}catch(Exception ex){
			EhiLogger.logError(ex);
			throw ex;
		}
	}
	private void mapFeatureType(ArrayList geomAttrAccu,IFMEFeature ret,ViewableWrapper wrapper,String attrNamePrefix)
	{
		boolean isStruct=attrNamePrefix!=null;
		if(!isStruct)attrNamePrefix="";
		//Viewable rootClass=getRoot(v);
		//if(rootClass==null){
		// rootClass=v;
		//}
		//EhiLogger.debug("aclass <"+v+">, root <"+rootClass+">");
		List attrv=wrapper.getAttrv();
		if(!isStruct){
			ret.setFeatureType(wrapper.getFmeFeatureType());
			ret.setStringAttribute(Ili2fme.XTF_ID,Ili2fme.ID_TYPE);
			ret.setStringAttribute(Ili2fme.XTF_BASKET,Ili2fme.ID_TYPE);
			ret.setStringAttribute(Ili2fme.XTF_GEOMATTR,Ili2fme.ILINAME_TYPE);
		}
		ret.setStringAttribute(attrNamePrefix+Ili2fme.XTF_CLASS,getIliQNameType());
		AttributeDef geomattr=null;
		for(int i=0;i<attrv.size();i++){
			ViewableTransferElement attro=(ViewableTransferElement)attrv.get(i);
			if(attro.obj instanceof AttributeDef){
				AttributeDef attr=(AttributeDef)attro.obj;
				Type type=attr.getDomainResolvingAliases();
				if (type instanceof PolylineType 
					|| type instanceof SurfaceOrAreaType
					|| type instanceof CoordType){
					if(formatMode==MODE_XTF){
						if(!isStruct){
							geomAttrAccu.add(attr);		
							if(geomattr==null){
								geomattr=attr;					
							}
						}else{
							EhiLogger.logAdaption(attr.getScopedName(null)+": Attribute ignored; FME supports only one geometry per feature");
						}
					}else if(formatMode==MODE_ITF){
						if(type instanceof SurfaceType && wrapper.getViewable()!=null){
							// surface attr in main table has no geometry
						}else{
							geomAttrAccu.add(attr);		
							if(geomattr==null){
								geomattr=attr;					
							}
						}
					}
				}else if(type instanceof CompositionType){
					Viewable structClass=((CompositionType)type).getComponentType();
					Viewable rootClass=ModelUtility.getRoot(structClass);
					if(rootClass==null){
					 rootClass=structClass;
					}
					mapFeatureType(geomAttrAccu,ret,(ViewableWrapper)fmeFeaturTypev.get(rootClass.getScopedName(null)),attr.getName()+"{}.");						
				}else{
					// TODO set attribute type
					ret.setStringAttribute(attrNamePrefix+attr.getName(),"xtf_char(20)");
				}
			}else if(attro.obj instanceof RoleDef){
				RoleDef role=(RoleDef)attro.obj;
				ret.setStringAttribute(attrNamePrefix+role.getName(),Ili2fme.ID_TYPE);
			}
		}
	}
	public boolean spatialEnabled() throws Exception {
		return false;
	}

	public void setConstraints(IFMEFeature arg0) throws Exception {
	}
	
	//
	// helpers
	//
	private void addCoord(IFMEFeature ret,IomObject value)
	throws InputDataException
	{
		if(value!=null){
			String c1=value.getattrvalue("C1");
			String c2=value.getattrvalue("C2");
			String c3=value.getattrvalue("C3");
			double xCoord;
			try{
				xCoord = Double.parseDouble(c1);
			}catch(Exception ex){
				throw new InputDataException("failed to read C1 <"+c1+">",ex);
			}
			double yCoord;
			try{
				yCoord = Double.parseDouble(c2);
			}catch(Exception ex){
				throw new InputDataException("failed to read C2 <"+c2+">",ex);
			}
			if(c3==null){
				ret.setDimension(IFMEFeature.FME_TWO_D);
				ret.add2DCoordinate(xCoord, yCoord);
			}else{
				double zCoord;
				try{
					zCoord = Double.parseDouble(c3);
				}catch(Exception ex){
					throw new InputDataException("failed to read C3 <"+c3+">",ex);
				}
				ret.setDimension(IFMEFeature.FME_THREE_D);
				ret.add3DCoordinate(xCoord, yCoord, zCoord);
			}
		}
	}
	private void addArc(IFMEFeature ret,IomObject value,double p)
	throws InputDataException
	{
		if(value!=null){
			String c1=value.getattrvalue("C1");
			String c2=value.getattrvalue("C2");
			String c3=value.getattrvalue("C3");
			String a1=value.getattrvalue("A1");
			String a2=value.getattrvalue("A2");
			double pt2_re;
			try{
				pt2_re = Double.parseDouble(c1);
			}catch(Exception ex){
				throw new InputDataException("failed to read C1 <"+c1+">",ex);
			}
			double pt2_ho;
			try{
				pt2_ho = Double.parseDouble(c2);
			}catch(Exception ex){
				throw new InputDataException("failed to read C2 <"+c2+">",ex);
			}
			double arcPt_re;
			try{
				arcPt_re = Double.parseDouble(a1);
			}catch(Exception ex){
				throw new InputDataException("failed to read A1 <"+a1+">",ex);
			}
			double arcPt_ho;
			try{
				arcPt_ho = Double.parseDouble(a2);
			}catch(Exception ex){
				throw new InputDataException("failed to read A2 <"+a2+">",ex);
			}
			int lastCoord=ret.numCoords();
			double p1[]=null;
			try{
				p1=ret.get2DCoordinate(lastCoord-1);
			}catch(FMEException ex)
			{
				EhiLogger.logError(ex);
				ret.add2DCoordinate(arcPt_re, arcPt_ho);
				ret.add2DCoordinate(pt2_re, pt2_ho);
			}
			double pt1_re=p1[0];
			double pt1_ho=p1[1];
			//EhiLogger.debug("pt1 "+pt1_re+", "+pt1_ho);
			//EhiLogger.debug("arc "+arcPt_re+", "+arcPt_ho);
			//EhiLogger.debug("pt2 "+pt2_re+", "+pt2_ho);
			/*
			if(c3==null){
				ret.setDimension(IFMEFeature.FME_TWO_D);
				ret.add2DCoordinate(p2_x, p2_y);
			}else{
				double zCoord = Double.parseDouble(c3);
				ret.setDimension(IFMEFeature.FME_THREE_D);
				ret.add3DCoordinate(p2_x, p2_y, zCoord);
			}
			*/
			// letzter Punkt ein Bogenzwischenpunkt?
		
			// Zwischenpunkte erzeugen

			// Distanz zwischen Bogenanfanspunkt und Zwischenpunkt
			double a=dist(pt1_re,pt1_ho,arcPt_re,arcPt_ho);
			// Distanz zwischen Zwischenpunkt und Bogenendpunkt 
			double b=dist(arcPt_re,arcPt_ho,pt2_re,pt2_ho);

			// Zwischenpunkte erzeugen, so dass maximale Pfeilhhe nicht 
			// berschritten wird
			// Distanz zwischen Bogenanfanspunkt und Bogenendpunkt 
			double c=dist(pt1_re,pt1_ho,pt2_re,pt2_ho);
			// Radius bestimmen
			double s=(a+b+c)/2.0;
			double ds=Math.atan2(pt2_re-arcPt_re,pt2_ho-arcPt_ho)-Math.atan2(pt1_re-arcPt_re,pt1_ho-arcPt_ho);
			double rSign=(Math.sin(ds)>0.0)?-1.0:1.0;
			double r=a*b*c/4.0/Math.sqrt(s*(s-a)*(s-b)*(s-c))*rSign;
			// Kreismittelpunkt
			double thetaM=Math.atan2(arcPt_re-pt1_re,arcPt_ho-pt1_ho)+Math.acos(a/2.0/r);
			double reM=pt1_re+r*Math.sin(thetaM);
			double hoM=pt1_ho+r*Math.cos(thetaM);

			// mindest Winkelschrittweite
			double theta=2*Math.acos(1-p/Math.abs(r));

			if(a>2*p){
				// Zentriwinkel zwischen pt1 und arcPt
				double alpha=2.0*Math.asin(a/2.0/Math.abs(r));
				// anzahl Schritte
				int alphan=(int)Math.ceil(alpha/theta);
				// Winkelschrittweite
				double alphai=alpha/(alphan*(r>0.0?1:-1));
				double ri=Math.atan2(pt1_re-reM,pt1_ho-hoM);
				for(int i=1;i<alphan;i++){
					ri += alphai;
					double pti_re=reM + Math.abs(r) * Math.sin(ri);
					double pti_ho=hoM + Math.abs(r) * Math.cos(ri);
					ret.add2DCoordinate(pti_re, pti_ho);
				}
			}

			ret.add2DCoordinate(arcPt_re, arcPt_ho);

			if(b>2*p){
				// Zentriwinkel zwischen arcPt und pt2
				double beta=2.0*Math.asin(b/2.0/Math.abs(r));
				// anzahl Schritte
				int betan=(int)Math.ceil((beta/theta));
				// Winkelschrittweite
				double betai=beta/(betan*(r>0.0?1:-1));
				double ri=Math.atan2(arcPt_re-reM,arcPt_ho-hoM);
				for(int i=1;i<betan;i++){
					ri += betai;
					double pti_re=reM + Math.abs(r) * Math.sin(ri);
					double pti_ho=hoM + Math.abs(r) * Math.cos(ri);
					ret.add2DCoordinate(pti_re, pti_ho);
				}
			}
			ret.add2DCoordinate(pt2_re, pt2_ho);
		}
	}
	private static double sqr(double x)
	{
		return x*x;
	}
	private static double dist(double re1,double ho1,double re2,double ho2)
	{
		double ret;
		ret=Math.sqrt(sqr(re2-re1)+sqr(ho2-ho1));
		return ret;
	}
	private void setPolyline(IFMEFeature ret,IomObject obj,boolean isSurfaceOrArea,double p)
	throws InputDataException
	{
		if(obj==null){
			return;
		}
		// is POLYLINE?
		if(!isSurfaceOrArea){
			ret.setGeometryType(IFMEFeature.FME_GEOM_LINE);
		}
		if(isSurfaceOrArea){
			IomObject lineattr=obj.getattrobj("lineattr",0);
			if(lineattr!=null){
				//writeAttrs(out,lineattr);
				EhiLogger.logAdaption("Lineattributes not supported by FME; ignored");							
			}
		}
		boolean clipped=obj.getobjectconsistency()==IomConstants.IOM_INCOMPLETE;
		for(int sequencei=0;sequencei<obj.getattrvaluecount("sequence");sequencei++){
			if(clipped){
				//out.startElement(tags::get_CLIPPED(),0,0);
			}else{
				// an unclipped polyline should have only one sequence element
				if(sequencei>0){
					EhiLogger.logError("unclipped polyline with multi 'sequence' elements");
					break;
				}
			}
			IomObject sequence=obj.getattrobj("sequence",sequencei);
			for(int segmenti=0;segmenti<sequence.getattrvaluecount("segment");segmenti++){
				IomObject segment=sequence.getattrobj("segment",segmenti);
				//EhiLogger.debug("segmenttag "+segment.getobjecttag());
				if(segment.getobjecttag().equals("COORD")){
					// COORD
					addCoord(ret,segment);
				}else if(segment.getobjecttag().equals("ARC")){
					// ARC
					addArc(ret,segment,p);
				}else{
					// custum line form
					EhiLogger.logAdaption("custom line form not supported by FME; ignored");
					//out.startElement(segment->getTag(),0,0);
					//writeAttrs(out,segment);
					//out.endElement(/*segment*/);
				}

			}
			if(clipped){
				//out.endElement(/*CLIPPED*/);
			}
		}
	}
	private void setSurface(IFMEFeature ret,IomObject obj,SurfaceOrAreaType type)
	throws InputDataException
	{
		if(obj==null){
			return;
		}
		ret.setGeometryType(IFMEFeature.FME_GEOM_DONUT);
		//IFMEFeatureVector bndries=session.createFeatureVector();
		boolean clipped=obj.getobjectconsistency()==IomConstants.IOM_INCOMPLETE;
		for(int surfacei=0;surfacei<obj.getattrvaluecount("surface");surfacei++){
			if(clipped){
				//out.startElement("CLIPPED",0,0);
			}else{
				// an unclipped surface should have only one surface element
				if(surfacei>0){
					EhiLogger.logError("unclipped surface with multi 'surface' elements");
					break;
				}
			}
			IomObject surface=obj.getattrobj("surface",surfacei);
			for(int boundaryi=0;boundaryi<surface.getattrvaluecount("boundary");boundaryi++){
				IomObject boundary=surface.getattrobj("boundary",boundaryi);
				//IFMEFeature fmeLine=session.createFeature();
				for(int polylinei=0;polylinei<boundary.getattrvaluecount("polyline");polylinei++){
					IomObject polyline=boundary.getattrobj("polyline",polylinei);
					setPolyline(ret,polyline,true,getP(type));
				}
				//bndries.append(fmeLine);
			}
			if(clipped){
				//out.endElement(/*CLIPPED*/);
			}
		}
		try{
			//int bndc=bndries.entries();
			//EhiLogger.debug("number of boundaries "+Integer.toString(bndc));
			//ret.makeDonuts(bndries,true);
			//fmeLog.logFeature(ret,IFMELogFile.FME_INFORM,-1);
			if(surfaceAsPolyline){
				ret.outerShell();
			}
		}catch(FMEException ex){
			EhiLogger.logError(ex);
		}
	}
	/** list of not yet processed struct values
	 */
	private ArrayList structQueue=null;
	private void enqueStructValue(int parentSqlId,String parentSqlType,String parentSqlAttr,IomObject struct,int structi)
	{
		structQueue.add(new StructWrapper(parentSqlId,parentSqlType,parentSqlAttr,struct,structi));
	}	
	private boolean doSetupIom=true;
	private void setupIom(){
		// setup iom
		if(doSetupIom){
			Ili2fme.setupIom();
			doSetupIom=false;
		}
	}
	private FmeLogListener listener=null;
	private void cleanup(){
		//EhiLogger.debug("cleanup");
		if(ioxReader!=null){
			try{
				ioxReader.close();
			}catch(IoxException ex){
				EhiLogger.logError(ex);
			}
			ioxReader=null;
		}
		if(inputFile!=null){
			try{
				inputFile.close();
			}catch(java.io.IOException ex){
				EhiLogger.logError(ex);
			}
			inputFile=null;
		}
		// did setup iom?
		if(!doSetupIom){
			//EhiLogger.debug("call Iom.end()");
			Ili2fme.endIom();
			doSetupIom=true;
		}
		if(listener!=null){
			Ili2fme.endLogging();
			listener=null;
		}

	}
	private String getIliQNameType()
	{
		return "xtf_char("+Integer.toString(iliQNameSize)+")";
	}
}
