package ch.interlis.iom_j.itf;
import ch.interlis.iox_j.*;
import ch.interlis.iox.IoxException;
import ch.interlis.iom.*;
import ch.ehi.basics.logging.EhiLogger;
import ch.interlis.ili2c.metamodel.*;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.Iterator;
import java.io.IOException;
import ch.interlis.iom_j.Iom_jObject;

/**
 * @author ce
 * @version $Revision: 1.0 $ $Date: 17.07.2006 $
 */
public class ItfWriter implements ch.interlis.iox.IoxWriter {
	private String modelName=null;
	private String topicName=null;
	private String className=null;
	private TransferDescription td=null;
	private HashMap tag2class=null;
	private int state=0;
	private ItfStartTransferEvent itfStart=null;
	private char blankCode='_';
	private ItfRawWriter out=null;
	public ItfWriter(java.io.OutputStream out,TransferDescription td)
	throws java.io.UnsupportedEncodingException
	,java.io.IOException
	{
		this.td=td;
		this.out=new ItfRawWriter(out,td.getIli1Format());
		blankCode=ModelUtilities.code2string(td.getIli1Format().blankCode).charAt(0);
	}
	public void close()
		throws IoxException 
	{
		flush();
		if(out!=null){
			out.close();
			out=null;
		}
		td=null;
		state=0;
		tag2class=null;
		itfStart=null;
	}
	public void flush()
		throws IoxException 
	{
		if(out!=null){
			out.flush();
		}
	}

	public void write(ch.interlis.iox.IoxEvent event) 
		throws IoxException 
	{
		if(event instanceof StartTransferEvent){
			if(state!=0){
				throw new IllegalStateException();
			}
			StartTransferEvent e=(StartTransferEvent)event;
			// SCNT
			try{
				out.writeRawText("SCNT");
				out.writeNewline();
				out.writeExplanation(e.getComment());
				out.writeRawText("////");
				out.writeNewline();
			}catch(IOException ex){
				throw new IoxException("failed to write SCNT section",ex);
			}
			// special kind (ITF specific) of start event?
			if(event instanceof ItfStartTransferEvent){
				itfStart=(ItfStartTransferEvent)event;
			}
			state=19;
		}else if(event instanceof StartBasketEvent){
			StartBasketEvent e=(StartBasketEvent)event;
			// first basket?
			if(state==19){
				if(itfStart!=null){
					// MTID/MOTR
					if(itfStart.getModelDefinition()!=null){
						try{
							out.writeRawText("MOTR");
							out.writeNewline();
							out.writeExplanation(itfStart.getModelDefinition());
							out.writeRawText("////");
							out.writeNewline();
						}catch(IOException ex){
							throw new IoxException("failed to write MOTR section",ex);
						}
					}else if(itfStart.getModelId()!=null){
						try{
							out.writeRawText("MTID ");
							out.writeRawText(itfStart.getModelId());
							out.writeNewline();
						}catch(IOException ex){
							throw new IoxException("failed to write MTID line",ex);
						}
					}else{
						throw new IllegalArgumentException();
					}
					String model[]=e.getType().split("\\.");
					modelName=model[0];
					// MODL
					try{
						out.writeRawText("MODL ");
						out.writeRawText(modelName);
						out.writeNewline();
					}catch(IOException ex){
						throw new IoxException("failed to write MODL line",ex);
					}
				}else{
					String model[]=e.getType().split("\\.");
					modelName=model[0];
					// MTID/MOTR
					try{
						out.writeRawText("MTID ");
						out.writeRawText(modelName);
						out.writeNewline();
					}catch(IOException ex){
						throw new IoxException("failed to write MTID line",ex);
					}
					// MODL
					try{
						out.writeRawText("MODL ");
						out.writeRawText(modelName);
						out.writeNewline();
					}catch(IOException ex){
						throw new IoxException("failed to write MODL line",ex);
					}
				}
				state=20;
			}
			if(state!=20){
				throw new IllegalStateException();
			}
			String topic[]=e.getType().split("\\.");
			if(!topic[0].equals(modelName)){
				EhiLogger.logAdaption("unexpected model name <"+topic[0]+">; ignored");
			}
			topicName=topic[1]; // topic[0] is name of model
			// TOPI
			try{
				out.writeRawText("TOPI ");
				out.writeRawText(topicName);
				out.writeNewline();
			}catch(IOException ex){
				throw new IoxException("failed to write TOPI line",ex);
			}
			// setup trigger for first object of table 
			className=null;
			state=30;
		}else if(event instanceof ObjectEvent){
			if(state!=30){
				throw new IllegalStateException();
			}
			ObjectEvent e=(ObjectEvent)event;
			String classv[]=e.getIomObject().getobjecttag().split("\\.");
			if(!classv[0].equals(modelName)){
				EhiLogger.logAdaption("unexpected model name <"+classv[0]+">; ignored");
			}
			if(!classv[1].equals(topicName)){
				EhiLogger.logAdaption("unexpected topic name <"+classv[1]+">; ignored");
			}
			// first object of a table?
			if(className==null || !className.equals(classv[2])){
				// not first table?
				if(className!=null){
					// ETAB
					try{
						out.writeRawText("ETAB");
						out.writeNewline();
					}catch(IOException ex){
						throw new IoxException("failed to write ETAB line",ex);
					}
				}
				className=classv[2];
				// TABL
				try{
					out.writeRawText("TABL ");
					out.writeRawText(className);
					out.writeNewline();
				}catch(IOException ex){
					throw new IoxException("failed to write TABL line",ex);
				}
			}
			// OBJE
			writeObject(e.getIomObject());
		}else if(event instanceof EndBasketEvent){
			if(state!=30){
				throw new IllegalStateException();
			}
			// if end of table
			if(className!=null){
				// ETAB
				try{
					out.writeRawText("ETAB");
					out.writeNewline();
				}catch(IOException ex){
					throw new IoxException("failed to write ETAB line",ex);
				}
			}
			// ETOP
			try{
				out.writeRawText("ETOP");
				out.writeNewline();
			}catch(IOException ex){
				throw new IoxException("failed to write ETOP line",ex);
			}
			state=20;
		}else if(event instanceof EndTransferEvent){
			if(state!=19 && state!=20){
				throw new IllegalStateException();
			}
			// at least one Start/EndBasket received?
			if(state==20){
				try{
					out.writeRawText("EMOD");
					out.writeNewline();
				}catch(IOException ex){
					throw new IoxException("failed to write EMOD line",ex);
				}
			}
			try{
				out.writeRawText("ENDE");
				out.writeNewline();
			}catch(IOException ex){
				throw new IoxException("failed to write ENDE line",ex);
			}
		}
		
	}
	private void writeObject(IomObject iomObj)
	throws IoxException
	{
		try{
			// write primattr
			if(td!=null){
				String iliQName=iomObj.getobjecttag();
				if(tag2class==null){
					tag2class=ModelUtilities.getTagMap(td);
				}
				if(!tag2class.containsKey(iliQName)){
					throw new IoxException("unknown class <"+iliQName+">");
				}
				out.writeRawText("OBJE");
				out.writeTid(iomObj.getobjectoid());
				Object aclassObj=tag2class.get(iliQName);
				AbstractClassDef aclass;
				AttributeDef attrOfHelperTab=null;
				ArrayList linev=null;
				if(aclassObj instanceof AbstractClassDef){
					// main table
					aclass=(AbstractClassDef)aclassObj;
					linev=ModelUtilities.getPolylineAttrs(aclass);
				}else{
					// SURFACE or AREA helper table
					attrOfHelperTab=(AttributeDef)aclassObj;
					aclass=(AbstractClassDef)attrOfHelperTab.getContainer();
					linev=new ArrayList();
					linev.add(attrOfHelperTab);
				}
				// area or surface table?
				if(attrOfHelperTab!=null){
					// if SURFACE add REF to main table
					if(attrOfHelperTab.getDomainResolvingAliases() instanceof SurfaceType){
						IomObject structvalue=iomObj.getattrobj(aclass.getName(),0);
						String ref=structvalue.getobjectrefoid();
						out.writeTid(ref);
					}
					// TODO line attributes
				}else{
					ArrayList attrv=ModelUtilities.getIli1AttrList(aclass);
					Iterator attri=attrv.iterator();
					while (attri.hasNext ()){
					  ViewableTransferElement obj = (ViewableTransferElement)attri.next();
					  if (obj.obj instanceof AttributeDef){
						AttributeDef attr = (AttributeDef) obj.obj;
						Type type = Type.findReal (attr.getDomain());
						if((type instanceof LineType)
							&& !(type instanceof AreaType)){
							// LineType's but not AreaType
							; // not part of main record
						}else{
						  // AreaType: centroid is part of main record
						  // other non LineType's
						  // attr.getName()
						  if(type instanceof AreaType){
							  Domain controlPointDomain = ((AreaType)type).getControlPointDomain();
							  CoordType coordType=(CoordType) Type.findReal (controlPointDomain.getType());
							  boolean is3D=coordType.getDimensions().length==3;
							  IomObject coord=iomObj.getattrobj(attr.getName(),0);
								writeNum(coord!=null ? coord.getattrvalue("C1") : null,coordType.getDimensions()[0]);
								writeNum(coord!=null ? coord.getattrvalue("C2") : null,coordType.getDimensions()[1]);
								if(is3D){
									writeNum(coord!=null ? coord.getattrvalue("C3") : null,coordType.getDimensions()[2]);
								}
						  }else if(type instanceof CoordType){
						  CoordType coordType=(CoordType)type;
							  boolean is3D=coordType.getDimensions().length==3;
							IomObject coord=iomObj.getattrobj(attr.getName(),0);
							writeNum(coord!=null ? coord.getattrvalue("C1") : null,coordType.getDimensions()[0]);
							writeNum(coord!=null ? coord.getattrvalue("C2") : null,coordType.getDimensions()[1]);
							if(is3D){
								writeNum(coord!=null ? coord.getattrvalue("C3") : null,coordType.getDimensions()[2]);
							}
						  }else if(type instanceof EnumerationType){
							String enumQName=iomObj.getattrvalue(attr.getName());
							  if(enumQName!=null){
								  // map ili-qname to itf-code 
								  String itfCode=mapXtfCode2ItfCode((EnumerationType)type,enumQName);
								  if(itfCode==null){
									  EhiLogger.logAdaption(iomObj.getobjecttag()+" "+iomObj.getobjectoid()+": unexpected code <"+enumQName+"> for attribute "+attr.getName()+"; written without mapping");
									itfCode=enumQName;
								  }
								out.writeValue(itfCode,false,getFieldSize(type));
							  }else{
								out.writeValue(null,false,getFieldSize(type));
							  }
						  }else if(type instanceof TextType){
								String txt=iomObj.getattrvalue(attr.getName());
								if(txt!=null){
									txt=txt.replace(' ',blankCode);
									out.writeValue(txt,false,((TextType)type).getMaxLength());
								}else{
									out.writeValue(null,false,((TextType)type).getMaxLength());
								}
						  }else if(type instanceof ReferenceType){
							String ref=iomObj.getattrvalue(attr.getName());
							out.writeTid(ref);
						  }else{
							String txt=iomObj.getattrvalue(attr.getName());
							out.writeValue(txt,false,getFieldSize(type));
						  }
						}
					  }else if(obj.obj instanceof RoleDef){
						  RoleDef role = (RoleDef) obj.obj;
						  if(!obj.embedded){
							IomObject structvalue=iomObj.getattrobj(role.getName(),0);
							String ref=structvalue!=null ? structvalue.getobjectrefoid() : null;
							out.writeTid(ref);
						  }else{
							// +oppend.getName()+" ->"+oppend.getDestination().getName());
							  IomObject structvalue=iomObj.getattrobj(role.getName(),0);
							  String ref=structvalue!=null ? structvalue.getobjectrefoid() : null;
							out.writeTid(ref);
						  }
					  }
					}
				}
				out.writeNewline();
				// linev
				if(linev!=null){
					Iterator linei=linev.iterator();
					while(linei.hasNext()){
						AttributeDef polyAttr=(AttributeDef)linei.next();
						Type type = Type.findReal (polyAttr.getDomain());
						Domain controlPointDomain = ((LineType)type).getControlPointDomain();
						CoordType coordType=(CoordType) Type.findReal (controlPointDomain.getType());
						IomObject value=iomObj.getattrobj(polyAttr.getName(),0);
						if(value!=null){
							writePolyline(value,coordType);
						}
						out.writeRawText("ELIN");
						out.writeNewline();
					}
				}

			}else{
				out.writeRawText("OBJE");
				out.writeTid(iomObj.getobjectoid());
				out.writeNewline();
			}
		}catch(IOException ex){
			throw new IoxException("failed to write OBJE line",ex);
		}
	}
	private void writeNum(String numValue,NumericalType type1)
	throws java.io.IOException
	{
		NumericType type=(NumericType)type1;
		if(numValue==null || numValue.length()==0){
			out.writeValue(null,true,getFieldSize(type));
			return;
		}
		int scalePos=numValue.toLowerCase().indexOf('e');
		if(scalePos>0){
			numValue=numValue.substring(0,scalePos);
		}
		int accuracy=type.getMinimum().getAccuracy();
		if(accuracy==0){
			if(numValue.indexOf('.')>0){
				numValue=numValue.substring(0,numValue.indexOf('.'));
			}
		}else{
			int decPointPos=numValue.indexOf('.');
			if(decPointPos==-1){
				numValue=numValue+"."+ch.ehi.basics.tools.StringUtility.STRING(accuracy,'0');
			}else{
				int cl=numValue.length()-decPointPos-1;
				if(cl>accuracy){
					numValue=numValue.substring(0,decPointPos+1+accuracy);
				}else{
					numValue=numValue+ch.ehi.basics.tools.StringUtility.STRING(accuracy-cl,'0');
				}
			}
		}
		out.writeValue(numValue,true,getFieldSize(type));
	}
	private HashMap typeCache=new HashMap();
	private int getFieldSize(Type type)
	{
		if(typeCache.containsKey(type)){
			return ((Integer)typeCache.get(type)).intValue();
		}
		int size;
		if(type instanceof EnumerationType){
			ArrayList accu=new ArrayList();
			size=Integer.toString(ch.interlis.ili2c.generator.Interlis1Generator.countEnumLeafs(((EnumerationType)type).getEnumeration())).length();
		}else if(type instanceof NumericType){
			NumericType numType=(NumericType)type;
			String fmt=ch.interlis.ili2c.generator.Interlis1Generator.genFmtField(0,numType.getMinimum(),numType.getMaximum());
			size=fmt.length();			
		}else{
			throw new IllegalArgumentException("unknown type");
		}
		typeCache.put(type,new Integer(size));
		return size;
	}

	private void writePolyline(IomObject obj,CoordType coordType)
	throws java.io.IOException
	{
		if(obj==null){
			return;
		}
		boolean is3D=coordType.getDimensions().length==3;
		boolean clipped=obj.getobjectconsistency()==IomConstants.IOM_INCOMPLETE;
		if(clipped){
			throw new IllegalArgumentException("IOM_INCOMPLETE not supported by ITF file format");
		}
		for(int sequencei=0;sequencei<obj.getattrvaluecount("sequence");sequencei++){
			if(clipped){
				//out.startElement(tags::get_CLIPPED(),0,0);
			}else{
				// an unclipped polyline should have only one sequence element
				if(sequencei>0){
					EhiLogger.logError("unclipped polyline with multi 'sequence' elements");
					break;
				}
			}
			IomObject sequence=obj.getattrobj("sequence",sequencei);
			for(int segmenti=0;segmenti<sequence.getattrvaluecount("segment");segmenti++){
				IomObject segment=sequence.getattrobj("segment",segmenti);
				if(segment.getobjecttag().equals("COORD")){
					// COORD
					if(segmenti==0){
						out.writeRawText("STPT");
					}else{
						out.writeRawText("LIPT");
					}
					writeNum(segment.getattrvalue("C1") ,coordType.getDimensions()[0]);
					writeNum(segment.getattrvalue("C2") ,coordType.getDimensions()[1]);
					if(is3D){
						writeNum(segment.getattrvalue("C3") ,coordType.getDimensions()[2]);
					}
					out.writeNewline();
				}else if(segment.getobjecttag().equals("ARC")){
					// ARC
					out.writeRawText("ARCP");
					writeNum(segment.getattrvalue("A1") ,coordType.getDimensions()[0]);
					writeNum(segment.getattrvalue("A2") ,coordType.getDimensions()[1]);
					//if(is3D){
					//	writeNum(segment.getattrvalue("A3") ,coordType.getDimensions()[2]);
					//}
					out.writeNewline();
					out.writeRawText("LIPT");
					writeNum(segment.getattrvalue("C1") ,coordType.getDimensions()[0]);
					writeNum(segment.getattrvalue("C2") ,coordType.getDimensions()[1]);
					if(is3D){
						writeNum(segment.getattrvalue("C3") ,coordType.getDimensions()[2]);
					}
					out.writeNewline();
				}else{
					// custum line form
					EhiLogger.logAdaption("custom line form not supported; ignored");
					//out.startElement(segment->getTag(),0,0);
					//writeAttrs(out,segment);
					//out.endElement(/*segment*/);
				}

			}
			if(clipped){
				//out.endElement(/*CLIPPED*/);
			}
		}
	}
	private HashMap enumTypes=new HashMap();
	private String mapXtfCode2ItfCode(EnumerationType type,String xtfCode)
	{
		HashMap mapping=null;
		if(!enumTypes.containsKey(type))
		{
			java.util.ArrayList ev=new java.util.ArrayList();
			ModelUtilities.buildEnumList(ev,"",type.getConsolidatedEnumeration());
			Iterator iter=ev.iterator();
			mapping=new HashMap();
			int code=0;
			while(iter.hasNext()){
				String ili1code=Integer.toString(code);
				String ili2code=(String)iter.next();
				mapping.put(ili2code,ili1code);
				code++;
			}
			enumTypes.put(type,mapping);
		}else{
			mapping=(HashMap)enumTypes.get(type);
		}
		if(mapping.containsKey(xtfCode)){
			return (String)mapping.get(xtfCode);
		}
		return null;
	}
	
	public static void main(String[] args) {
		try{
			java.io.FileOutputStream out=new java.io.FileOutputStream(args[0]);
			ch.interlis.iox.IoxWriter writer=new ItfWriter(out,null);
			StartTransferEvent start=new StartTransferEvent();
			start.setComment("kleiner Kommentar");
			writer.write(start);
			//...
			// while(more baskets){
			//	writer.write(StartBasketEvent);
			StartBasketEvent startBasket=new StartBasketEvent("TestModel.TestTopic","1");
			writer.write(startBasket);
			//	...
			//	while(more objects){
			// 		...
			//		writer.write(ObjectEvent);
			IomObject iomObj=new Iom_jObject("TestModel.TestTopic.TestClassA","1");
			ObjectEvent o=new ObjectEvent(iomObj);
			writer.write(o);
			iomObj.setobjectoid("2");
			writer.write(o);
			
			iomObj.setobjecttag("TestModel.TestTopic.TestClassB");
			iomObj.setobjectoid("3");
			writer.write(o);
			
			// 		...
			//	}
			//		...
			//	writer.write(EndBasketEvent);
			writer.write(new EndBasketEvent());
			// }
			writer.write(new EndTransferEvent());
			writer.close();
			out.close();
		}catch(Exception ex)
		{
			EhiLogger.logError(ex);
		}
		
	}
}
